#include <CGAL/basic.h>

#include <CGAL/CORE_BigInt.h>                      //NT
#include <CGAL/Algebraic_kernel_d_1.h>             //Algebraic Kernel
#include <CGAL/Arr_rational_function_traits_2.h>   //Traits
#include <CGAL/Arrangement_2.h>                    //Arrangement
#include <CGAL/Sweep_line_2_algorithms.h>
#include <boost/foreach.hpp>

#include <CGAL/Algebraic_kernel_2_1.h>

#if 0
typedef CORE::BigInt                               NT;
typedef CGAL::Algebraic_kernel_d_1<NT>             AK1; 
#else
typedef CORE::BigRat                               Rational;
//typedef Rational                                   FT;
typedef CGAL::Lazy_exact_nt<Rational>              FT; 
typedef CGAL::Algebraic_kernel_2_1<FT>             AK1;
#endif
typedef CGAL::Arr_rational_function_traits_2<AK1>  Traits_2;
typedef CGAL::Arrangement_2<Traits_2>              Arrangement_2;

int main(int argc, char* argv[])
{

  // testing that all types are present: 
  
  // Traits 
  typedef Traits_2::Polynomial_1         Polynomial_1;
  typedef Traits_2::Algebraic_real_1     Algebraic_real_1;
  typedef Traits_2::Coefficient          Coefficient; 
  typedef Traits_2::Bound                Bound; 
  typedef Traits_2::Algebraic_kernel_d_1 Algebraic_kernel_d_1; 
  typedef Traits_2::Construct_curve_2 Construct_curve_2;
  typedef Traits_2::Construct_x_monotone_curve_2 Construct_x_monotone_curve_2; 
  typedef Traits_2::Construct_point_2 Construct_point_2;
  
  // typedef induced by concept 

  typedef Traits_2::Curve_2 Curve_2;
  typedef Traits_2::X_monotone_curve_2 X_monotone_curve_2; 
  typedef Traits_2::Point_2 Point_2; 
  
  typedef Traits_2::Compare_x_2 Compare_x_2;
  typedef Traits_2::Compare_xy_2 Compare_xy_2;
  typedef Traits_2::Construct_min_vertex_2 Construct_min_vertex_2;
  typedef Traits_2::Construct_max_vertex_2 Construct_max_vertex_2; 
  typedef Traits_2::Is_vertical_2 Is_vertical_2;
  typedef Traits_2::Compare_y_at_x_2 Compare_y_at_x_2;
  typedef Traits_2::Compare_y_at_x_left_2 Compare_y_at_x_left_2;
  typedef Traits_2::Compare_y_at_x_right_2 Compare_y_at_x_right_2;
  typedef Traits_2::Equal_2 Equal_2; 
  typedef Traits_2::Parameter_space_in_x_2 Parameter_space_in_x_2; 
  typedef Traits_2::Parameter_space_in_y_2 Parameter_space_in_y_2;
  typedef Traits_2::Compare_x_at_limit_2 Compare_x_at_limit_2;
  typedef Traits_2::Compare_x_near_limit_2 Compare_x_near_limit_2; 
  typedef Traits_2::Compare_y_near_boundary_2 Compare_y_near_boundary_2; 
  typedef Traits_2::Intersect_2 Intersect_2; 
  typedef Traits_2::Split_2 Split_2; 
  typedef Traits_2::Are_mergeable_2 Are_mergeable_2; 
  typedef Traits_2::Merge_2 Merge_2;
  typedef Traits_2::Make_x_monotone_2 Make_x_monotone_2;

  // construction traits 
  // default construction 
  {
    Traits_2 traits; 
    const Algebraic_kernel_d_1* ak (traits.algebraic_kernel_d_1()); (void) ak;
  }
  
  
  // construction from pointer to AK 
  Algebraic_kernel_d_1 ak; 
  Traits_2 traits(&ak); 
  
  
  traits.cleanup_cache(); 

  Construct_curve_2 construct_curve_2 = traits.construct_curve_2_object(); 
  
  Polynomial_1 P(0);
  Polynomial_1 Q(1);
  Algebraic_real_1 one(1);
  Algebraic_real_1 two(2);
  {
    typedef Construct_curve_2::Polynomial_1 Polynomial_1;
    typedef Construct_curve_2::Algebraic_real_1 Algebraic_real_1; 
    typedef Construct_curve_2::result_type Curve_2; 
    typedef Construct_curve_2::argument_type argument_type;
    typedef Construct_curve_2::first_argument_type first_argument_type; 
    typedef Construct_curve_2::second_argument_type second_argument_type; 
  }
  {Curve_2 curve = construct_curve_2(P);}
  {Curve_2 curve = construct_curve_2(P,one,true);}
  {Curve_2 curve = construct_curve_2(P,one,false);}
  {Curve_2 curve = construct_curve_2(P,one,two);}
  {Curve_2 curve = construct_curve_2(P,Q);}
  {Curve_2 curve = construct_curve_2(P,Q,one,true);}
  {Curve_2 curve = construct_curve_2(P,Q,one,false);}
  {Curve_2 curve = construct_curve_2(P,Q,one,two);}
  {Curve_2 curve = construct_curve_2(P.begin(),P.end());}
  {Curve_2 curve = construct_curve_2(P.begin(),P.end(),one,true);}
  {Curve_2 curve = construct_curve_2(P.begin(),P.end(),one,false);}
  {Curve_2 curve = construct_curve_2(P.begin(),P.end(),one,two);}
  {Curve_2 curve = construct_curve_2(P.begin(),P.end(),Q.begin(),Q.end());}
  {Curve_2 curve = construct_curve_2(P.begin(),P.end(),Q.begin(),Q.end(),one,true);}
  {Curve_2 curve = construct_curve_2(P.begin(),P.end(),Q.begin(),Q.end(),one,false);}
  {Curve_2 curve = construct_curve_2(P.begin(),P.end(),Q.begin(),Q.end(),one,two);}

  
  Construct_x_monotone_curve_2 
    construct_x_monotone_curve_2 = traits.construct_x_monotone_curve_2_object();
  {
    typedef Construct_x_monotone_curve_2::Polynomial_1 Polynomial_1;
    typedef Construct_x_monotone_curve_2::Algebraic_real_1 Algebraic_real_1; 
    typedef Construct_x_monotone_curve_2::X_monotone_curve_2 X_monotone_curve_2; 
    typedef Construct_x_monotone_curve_2::argument_type argument_type;
    typedef Construct_x_monotone_curve_2::first_argument_type first_argument_type; 
    typedef Construct_x_monotone_curve_2::second_argument_type second_argument_type; 
  }
 
  {X_monotone_curve_2 x_monotone_curve = construct_x_monotone_curve_2(P);}
  {X_monotone_curve_2 x_monotone_curve = construct_x_monotone_curve_2(P,one,true);}
  {X_monotone_curve_2 x_monotone_curve = construct_x_monotone_curve_2(P,one,false);}
  {X_monotone_curve_2 x_monotone_curve = construct_x_monotone_curve_2(P,one,two);}
  {X_monotone_curve_2 x_monotone_curve = construct_x_monotone_curve_2(P,Q);}
  {X_monotone_curve_2 x_monotone_curve = construct_x_monotone_curve_2(P,Q,one,true);}
  {X_monotone_curve_2 x_monotone_curve = construct_x_monotone_curve_2(P,Q,one,false);}
  {X_monotone_curve_2 x_monotone_curve = construct_x_monotone_curve_2(P,Q,one,two);}
  {X_monotone_curve_2 x_monotone_curve = construct_x_monotone_curve_2(P.begin(),P.end());}
  {X_monotone_curve_2 x_monotone_curve = construct_x_monotone_curve_2(P.begin(),P.end(),one,true);}
  {X_monotone_curve_2 x_monotone_curve = construct_x_monotone_curve_2(P.begin(),P.end(),one,false);}
  {X_monotone_curve_2 x_monotone_curve = construct_x_monotone_curve_2(P.begin(),P.end(),one,two);}
  {X_monotone_curve_2 x_monotone_curve = construct_x_monotone_curve_2(P.begin(),P.end());}
  {X_monotone_curve_2 x_monotone_curve = construct_x_monotone_curve_2(P.begin(),P.end(),Q.begin(),Q.end(),one,true);}
  {X_monotone_curve_2 x_monotone_curve = construct_x_monotone_curve_2(P.begin(),P.end(),Q.begin(),Q.end(),one,false);}
  {X_monotone_curve_2 x_monotone_curve = construct_x_monotone_curve_2(P.begin(),P.end(),Q.begin(),Q.end(),one,two);}

  {
    typedef Curve_2::Polynomial_1 Polynomial_1; 
    typedef Curve_2::Algebraic_real_1 Algebraic_real_1;
  }

  {
     
    Curve_2 curve = construct_curve_2(P,Q);
    {Curve_2 dummy;}
    {Curve_2 dummy(curve);}
    {Curve_2 dummy = curve;}
    
    assert(CGAL::degree(curve.numerator())  >=0);
    assert(CGAL::degree(curve.denominator())>=0);
    assert(curve.is_continuous());
    assert(curve.left_parameter_space_in_x()==CGAL::ARR_LEFT_BOUNDARY);
    assert(curve.right_parameter_space_in_x()==CGAL::ARR_RIGHT_BOUNDARY);
  }
  {
    Curve_2 curve = construct_curve_2(P,one,two);
    assert(CGAL::degree(curve.numerator())  >=0);
    assert(CGAL::degree(curve.denominator())>=0);
    assert(curve.is_continuous());
    assert(curve.left_parameter_space_in_x()==CGAL::ARR_INTERIOR);
    assert(curve.right_parameter_space_in_x()==CGAL::ARR_INTERIOR);
    assert(one == curve.left_x()); 
    assert(two == curve.right_x());
  }
  // X_monotone_curve_2
  {
    typedef X_monotone_curve_2::Polynomial_1 Polynomial_1;
    typedef X_monotone_curve_2::Algebraic_real_1 Algebraic_real_1; 
    typedef X_monotone_curve_2::Point_2 Point_2; 
    
    
    X_monotone_curve_2 xcurve= construct_x_monotone_curve_2(P,Q,one,two);

    {X_monotone_curve_2 dummy;}
    {X_monotone_curve_2 dummy(xcurve);}
    {X_monotone_curve_2 dummy = xcurve;}
        
    assert(CGAL::degree(xcurve.numerator())  >=0);
    assert(CGAL::degree(xcurve.denominator())>=0);
    assert(xcurve.source_parameter_space_in_x()==CGAL::ARR_INTERIOR);
    assert(xcurve.source_parameter_space_in_y()==CGAL::ARR_INTERIOR);
    xcurve.source(); 
    xcurve.target(); 
    xcurve.left(); 
    xcurve.right();
    xcurve.source_x(); 
    xcurve.target_x(); 
    xcurve.left_x(); 
    xcurve.right_x();
    assert(xcurve.is_left_to_right()); 
  }


  // Point_2
  {
    typedef Point_2::Polynomial_1 Polynomial_1;
    typedef Point_2::Algebraic_real_1 Algebraic_real_1;
    typedef Point_2::Bound Bound; 


    {  
      Construct_point_2 construct_point_2 = traits.construct_point_2_object(); 
      Point_2 p = construct_point_2(Rational(1),Rational(1));
    }
    
    X_monotone_curve_2 xcurve= construct_x_monotone_curve_2(P,Q,one,two);
    Point_2 p = xcurve.left(); 
    Point_2 q = xcurve.right();
    
    p.numerator();
    p.denominator();
    p.to_double();
    p.x();
    p.y();
    p.approximate_absolute_x(1);
    p.approximate_absolute_y(1);
    p.approximate_relative_x(1);
    p.approximate_relative_y(1);
    
    {Point_2 dummy = p;}
    {Point_2 dummy(p);}
  }
  
  {
    Arrangement_2 arr; 
    const Traits_2* traits = arr.traits();(void) traits; 
  }

  { 
    
    const Traits_2 traits;  
    traits.cleanup_cache(); 
    assert( traits.cache().rat_func_map().size()==0);
    assert( traits.cache().rat_pair_map().size()==0);

    Construct_curve_2 construct_curve_2 = traits.construct_curve_2_object();
    
    Polynomial_1 x = CGAL::shift(Polynomial_1(1),1);
    
    {      
      std::vector<Curve_2> curves; 
      std::vector<X_monotone_curve_2> xcurves;
      std::vector<Point_2> points;
      curves.push_back(construct_curve_2(Polynomial_1(1)));
      curves.push_back(construct_curve_2(x-2,x-5));
      curves.push_back(construct_curve_2(5*x-5,x-4));
      curves.push_back(construct_curve_2(x-1,6*x-2));
      
     
      BOOST_FOREACH(const Curve_2& curve, curves){
        assert(CGAL::degree(curve.numerator())>=0);
      }   
      CGAL::compute_subcurves(curves.begin(),curves.end(),std::back_inserter(xcurves),false,traits); 
      BOOST_FOREACH(const X_monotone_curve_2& xcurve, xcurves){
        assert(CGAL::degree(xcurve.numerator())>=0);
      }    
	
      CGAL::compute_intersection_points (curves.begin(),curves.end(),std::back_inserter(points),false,traits);
      BOOST_FOREACH(const Point_2& point, points){
        assert(CGAL::degree(point.numerator())>=0);
       }  
    }  
    traits.cleanup_cache(); 
    assert( traits.cache().rat_func_map().size()==0);
    assert( traits.cache().rat_pair_map().size()==0);
    {      
      std::vector<Curve_2> curves; 
      std::vector<X_monotone_curve_2> xcurves;
      std::vector<Point_2> points;
      
      curves.push_back(construct_curve_2(Polynomial_1(1)));
      curves.push_back(construct_curve_2(x-2,x-5));
      curves.push_back(construct_curve_2(5*x-5,x-4));
      curves.push_back(construct_curve_2(x-1,6*x-2));
      
      traits.cleanup_cache();
      BOOST_FOREACH(const Curve_2& curve, curves){
        assert(CGAL::degree(curve.numerator())>=0);
      }   
      CGAL::compute_subcurves(curves.begin(),curves.end(),std::back_inserter(xcurves),false,traits); 
      BOOST_FOREACH(const X_monotone_curve_2& xcurve, xcurves){
        assert(CGAL::degree(xcurve.numerator())>=0);
      }    
      CGAL::compute_intersection_points (curves.begin(),curves.end(),std::back_inserter(points),false,traits);
      BOOST_FOREACH(const Point_2& point, points){
        assert(CGAL::degree(point.numerator())>=0);
      }  
      std::sort(points.begin(),points.end(), traits.compare_xy_2_object());
      std::sort(points.begin(),points.end(), traits.compare_x_2_object());
    }
  }

  return 0;
}
