#!/usr/bin/perl

use strict;
use Getopt::Long;

my $help = 0;
my $strict_vrml = 0;

my $width = 1;
my $depth = 1;
my $height = 1;
my $bounding_box = 0;

my $density = 16;

$Getopt::Long::ignorecase = 0;
GetOptions("help" => \$help,
	   "strict-vrml" => \$strict_vrml,
           "bounding-box" => \$bounding_box,
           "width" => \$width,
           "depth" => \$depth,
           "height" => \$height,
           "density" => \$density) or die "Bad command line option!";
usage() if $help;

sub usage {
    printf "gen_hyperbolic_paraboloid options";
    printf "
\t--help\t\t\tprint this help message
\t--strict-vrml\t\tgenerate standard VRML
\t--bounding-box\t\tgenerate the bounding box
\t--width <width>\t\tset the width to <width>
\t--depth <depth>\t\tset the depth to <depth>
\t--height <height>\tset the height to <height>
\t--density <num>\t\tset the density to <num>
";
  exit(0);
}

my $x_min = -$width/2;
my $x_max = $width/2;
my $y_min = -$depth/2;
my $y_max = $depth/2;
my $z_min = -$height/2;
my $z_max = $height/2;

print '#VRML V2.0 utf8',"\n";
if (!$strict_vrml) {
  print 'DEF CONFIG Configuration {
  accumulation Accumulation { enabled TRUE }
}
DEF BACKGROUND ColorBackground {
  color 1 1 1 0
  clearStencil TRUE
}
Viewpoint {
  type "ORTHOGONAL"
  fieldOfView 2.83
}
DEF SNAP_KEY SingleKeySensor { key "S" }
DEF SNAP Snapshot {
  image Image { }
  fileFormat "jpg"
  sequence FALSE
  dirName "."
  fileName "hyperbolic_paraboloid"
}',"\n";
}
print 'DEF EXAMINE_NAV NavigationInfo { type [ "EXAMINE" "ANY" ] }
DEF TRANSFORMER_NAV NavigationInfo { type [ "TRANSFORM" ] }
Transform {
  rotation 1 0 0 -1.30899 # -75
  children [
    Transform {
      rotation 0 0 1 -.261799 # -15
      children [',"\n";
if ($bounding_box) {
  print '        # The bounding box
        Shape {
          appearance Appearance {
            material Material { diffuseColor 0 0 0 }
          }
          geometry IndexedLineSet {
            coord Coordinate {
              point [',"\n";
  print "                     $x_min $y_min $z_min, $x_min $y_min $z_max,\n";
  print "                     $x_min $y_max $z_min, $x_min $y_max $z_max,\n";
  print "                     $x_max $y_min $z_min, $x_max $y_min $z_max,\n";
  print "                     $x_max $y_max $z_min, $x_max $y_max $z_max\n";
  print '                    ]
            }
            coordIndex [0 1, 0 2, 0 4, 1 3, 1 5, 2 3,
                        2 6, 3 7, 4 5, 4 6, 5 7, 6 7]
          }
        }',"\n";
}
print '        # The hyperbolic paraboloid
        # 1st rules
        Shape {
          appearance DEF RULE_APP Appearance {
            material Material { diffuseColor 0 0 0 }
          }
          geometry IndexedLineSet {
            coord Coordinate {
              point [',"\n";
my $y_src = $y_min;
my $y_trg = $y_max;
for (my $i = 0; $i < $density; ++$i) {
  my $x_src = $x_min + $width * $i / $density;
  my $z_src = $y_max - $height * $i / $density;
  my $x_trg = $x_src;
  my $z_trg = $z_min + $height * $i / $density;
  print "                     $x_src $y_src $z_src, $x_trg $y_trg $z_trg,\n";
}
print "                     $x_max $y_min $z_min, $x_max $y_max $z_max\n";
print '                    ]
            }
            coordIndex [',"\n";
for (my $i = 0; $i <= $density; ++$i) {
  my $source_id = 2 * $i;
  my $target_id = $source_id + 1;
  print "                        $source_id $target_id\n";
}
print '                       ]
          }
        }
        # 2nd rules
        Shape {
          appearance USE RULE_APP
          geometry IndexedLineSet {
            coord Coordinate {
              point [',"\n";
my $x_src = $x_min;
my $x_trg = $x_max;
for (my $i = 0; $i < $density; ++$i) {
  my $y_src = $y_min + $depth * $i / $density;
  my $z_src = $z_max - $height * $i / $density;
  my $y_trg = $y_src;
  my $z_trg = $z_min + $height * $i / $density;
  print "                     $x_src $y_src $z_src, $x_trg $y_trg $z_trg,\n";
}
print "                     $x_min $y_max $z_min, $x_max $y_max $z_max\n";
print '                    ]
            }
            coordIndex [',"\n";
for (my $i = 0; $i <= $density; ++$i) {
  my $source_id = 2 * $i;
  my $target_id = $source_id + 1;
  print "                        $source_id $target_id\n";
}
print '                       ]
          }
        }
        # Surface
        Shape {
          appearance Appearance {
            material Material {
              diffuseColor 0.7 0.5 0.6
              transparency 0.5
            }
          }
          geometry IndexedFaceSet {
            solid FALSE
            coord Coordinate {
              point [',"\n";
for (my $i = 0; $i < $density; ++$i) {
  my $x = -$width/2 + $width * $i / $density;
  my $z_src = $z_max - $height * $i / $density;
  my $z_trg = $z_min + $height * $i / $density;
  for (my $j = 0; $j < $density; ++$j) {
    my $y = $y_min + $depth * $j / $density;
    my $z = $z_src + ($z_trg - $z_src) * $j / $density;
    print "                     $x $y $z,\n";
  }
  print "                     $x $y_max $z_trg,\n";
}
for (my $j = 0; $j < $density; ++$j) {
  my $y = $y_min + $depth * $j / $density;
  my $z = $z_min + $height * $j / $density;
  print "                     $x_max $y $z,\n";
}
print "                     $x_max $y_max $z_max\n";
print '                    ]
            }
            coordIndex [',"\n";
for (my $i = 0; $i < $density; ++$i) {
  for (my $j = 0; $j < $density; ++$j) {
    my $v0 = $i * ($density + 1) + $j;
    my $v1 = $v0 + 1;
    my $v2 = ($i + 1) * ($density + 1) + $j;
    my $v3 = $v2 + 1;
    print "                        $v0 $v3 $v1 -1, $v0 $v2 $v3 -1\n";
  }
}
print '                       ]
          }
        }
        # The axes:
        DEF AXES_SWITCH Switch {
          whichChoice 0
          children [
            DEF AXES Transform {
              scale 0.1 0.1 0.1
              children [
                Transform {
                  translation 3.5 0 0
                  rotation 0 0 1 .261799 # 15
                  children [
                    Transform {
                      rotation 1 0 0 1.30899 # 75
                      translation 0 0 0
                      children [
                        Shape {
                          appearance DEF TEXT_APP Appearance {
                            material Material { diffuseColor 0 0 0 }
                          }
                          geometry Text {
                            string "X"
                            fontStyle FontStyle { size 0.5 }
                          }
                        }
                       ]
                    }
                   ]
                }
                Transform {
                  translation 0 3.5 0.4
                  rotation 0 0 1 .261799 # 15
                  children [
                    Transform {
                      rotation 1 0 0 1.30899 # 75
                      translation 0 0 -0.2
                      children [
                        Shape {
                          appearance USE TEXT_APP
                          geometry Text {
                            string "Y"
                            fontStyle FontStyle { size 0.5 }
                          }
                        }
                       ]
                    }
                   ]
                }
                Transform {
                  translation 0 0 3.5
                  rotation 0 0 1 .261799 # 15
                  children [
                    Transform {
                      rotation 1 0 0 1.30899 # 75
                      translation -0.1 -0.1 0
                      children [
                        Shape {
                          appearance DEF TEXT_APP Appearance {
                            material Material { diffuseColor 0 0 0 }
                          }
                          geometry Text {
                            string "Z"
                            fontStyle FontStyle { size 0.5 }
                          }
                        }
                       ]
                    }
                   ]
                }
                Shape {
                  priority 100
                  appearance DEF AXES_APP Appearance {
                    material Material { diffuseColor 0.5 0.5 0.5 }
                  }
                  geometry Sphere { radius 0.1 }
                }
                DEF CS_AXIS Transform {
                  translation 0 1.5 0
                  children [
                    Shape {
                      priority 100
                      appearance USE AXES_APP
                      geometry Cylinder {
                        radius 0.1
                        height 3
                        set_is_bottom_visible FALSE
                        set_is_top_visible FALSE
                      }
                    }
                    Transform {
                      translation 0 1.7 0
                      children [
                        Shape {
                          priority 100
                          appearance USE AXES_APP
                          geometry Cone {
                            bottomRadius 0.2
                            height 0.4
                          }
                        }
                       ]
                    }
                   ]
                }
                Transform {
                  rotation 0 0 -1 1.57
                  children [ USE CS_AXIS ]
                }
                Transform {
                  rotation 1 0 0 1.57
                  children [ USE CS_AXIS ]
                }
               ]
            }
           ]
        }
      ]
    }
   ]
}',"\n";
if (!$strict_vrml) {
  print 'ROUTE SNAP_KEY.state TO SNAP.trigger',"\n";
}

exit(0);
